#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <assert.h>
#include <netinet/in.h>
#include <pthread.h>
#include <ctype.h>

#ifdef _LINUX_
#include <stdint.h>
#endif /* _LINUX_ */

#ifdef _DARWIN_
#include <inttypes.h>
#endif /* _DARWIN_ */

#include <sys/time.h>
#include <sys/types.h>

#include "mysock_impl.h"
#include "packetdump.h"
#include "stcp_protocol.h"

/**
 * Open a dump file and initialize the file.
 */
FILE* dump_open(const char *fname, int thiszone, int snaplen);

/** 
 * Write data into the log file
 */
void dump(FILE *fp, const struct pcap_pkthdr *h, const unsigned char *sp);

/**
 * Close the file
 */
void dump_close(FILE *fp);

uint16_t cksum(uint16_t * buf, int bytecount)
{
  int count = bytecount / 2;
  register uint32_t sum = 0;
  while (count--) {
    sum += *buf++;
    if (sum & 0xffff0000) { // carry out case
      sum &= 0xFFFF;
      sum++;
    }
  }
  if (bytecount % 2 != 0) {	/* Handle odd byte with addition of low order 0 octet */
    sum += *buf & 0xff00;
    if (sum & 0xffff0000) { // carry out case
      sum &= 0xFFFF;
      sum++;
    }
  }
  return ~(sum & 0xFFFF);
}

int dumpbytes(FILE * stream, char * buffer, int count)
{
  const int bytesperline = 16;
  int lines = count / bytesperline;
  if (count % bytesperline != 0) lines++;
  int bytesout = 0;
  int asciiout = 0;
  int line, col;
  char * bptr = buffer;

  for (line = 0; line < lines; line++) {
    char * linestart = bptr;
    for (col = 0; col < bytesperline; col++) {
      if (bytesout >= count) break;
      fprintf(stream, "%2.2x ", (*bptr) & 0xff);
      bptr++;
      bytesout++;
    }
    for (; col < bytesperline; col++) {
      fprintf(stream, "   ");
    }
    fprintf(stream, ":: |");
    bptr = linestart;
    for (col = 0; col < bytesperline; col++) {
      if (asciiout >= count) break;
      if (isprint(*bptr)) {
	fprintf(stream, "%c", *bptr);
      } else {
	fprintf(stream, ".");
      }
      bptr++;
      asciiout++;
    }
    fprintf(stream, "|\n");
  }
  return bytesout;
}


/*-----------------------------------------------------------------------------
 * Method: log_packet()
 * Scope: Local 
 *
 *---------------------------------------------------------------------------*/

void log_packet(mysock_context_t * ctx, uint8_t* buf, int len, int direction)
{
    struct pcap_pkthdr h;
    int size;

    /* REQUIRES */
    assert(ctx);

    if(!ctx->logstream)
    {return; }

    PTHREAD_CALL(pthread_mutex_lock(&ctx->loglock));

    size = min(PACKET_DUMP_SIZE, len + sizeof(struct enet_hdr) + sizeof(struct ip_hdr));

    gettimeofday(&h.ts, 0);
    h.caplen = size;
    h.len = (size < PACKET_DUMP_SIZE) ? size : PACKET_DUMP_SIZE;
    ctx->ipptr->ip_len = htons(20 + len);
    ctx->ipptr->ip_sum = 0;
    if (direction == 0) { // local is source; remote is destination
      ctx->ipptr->ip_src.s_addr = ctx->localaddr;
      ctx->ipptr->ip_dst.s_addr = ctx->remoteaddr;
    } else {
      ctx->ipptr->ip_src.s_addr = ctx->remoteaddr;
      ctx->ipptr->ip_dst.s_addr = ctx->localaddr;
    }
    uint16_t mysum = cksum((uint16_t *)(ctx->ipptr), (ctx->ipptr->ip_hl) << 2);
    ctx->ipptr->ip_sum = mysum;
    dump(ctx->logstream, &h, (uint8_t *)ctx->enetptr, (uint8_t *)ctx->ipptr, buf);
    fflush(ctx->logstream);

    PTHREAD_CALL(pthread_mutex_unlock(&ctx->loglock));

} /* -- sr_log_packet -- */


static void
write_header(FILE *fp, int linktype, int thiszone, int snaplen)
{
        struct pcap_file_header hdr;

        hdr.magic = TCPDUMP_MAGIC;
        hdr.version_major = PCAP_VERSION_MAJOR;
        hdr.version_minor = PCAP_VERSION_MINOR;

        hdr.thiszone = thiszone;
        hdr.snaplen = snaplen;
        hdr.sigfigs = 0;
        hdr.linktype = linktype;

        if (fwrite((char *)&hdr, sizeof(hdr), 1, fp) != 1)
                fprintf(stderr, "sf_write_header: can't write header\n");
}

/*
 * Initialize so that sf_write_header() will output to the file named 'fname'.
 */
FILE *
dump_open(const char *fname, int thiszone, int snaplen)
{       
  FILE *fp;
 
        if (fname[0] == '-' && fname[1] == '\0')
                fp = stdout;
        else {
                fp = fopen(fname, "w");
                if (fp == NULL) {
                        fprintf(stderr, "sr_dump_open: can't open %s",
                            fname);
                        return (NULL);
                }
        }

        write_header(fp, LINKTYPE_ETHERNET, thiszone, snaplen);
	fflush(fp);

        return fp;
}

/*
 * Output a packet to the initialized dump file.
 */
void
dump(FILE *fp, const struct pcap_pkthdr *h, 
     uint8_t * enethdr, uint8_t * iphdr, const unsigned char *sp)
{
        struct pcap_sf_pkthdr sf_hdr;

        sf_hdr.ts.tv_sec  = h->ts.tv_sec;
        sf_hdr.ts.tv_usec = h->ts.tv_usec;
        sf_hdr.caplen     = h->caplen;
        sf_hdr.len        = h->len;

	int splen = h->caplen - (sizeof(struct enet_hdr) + sizeof(struct ip_hdr));
        /* XXX we should check the return status */
        (void)fwrite(&sf_hdr, sizeof(sf_hdr), 1, fp);
        (void)fwrite((char *)enethdr, sizeof(struct enet_hdr), 1, fp);
        (void)fwrite((char *)iphdr, sizeof(struct ip_hdr), 1, fp);
        (void)fwrite((char *)sp, splen, 1, fp);
}

void
dump_close(FILE *fp)
{
  fclose(fp);
}

